/***********************************************************************
*   $Id$
*   descr:      class CharString
*               garantee the allocation by "char *" strings.
*
*
*   author:     PINEDA Walter
*
*   $Log$
*
*   2 Aout 1995 : MF : add methods Append's
*                      changed delete sequence in Copy
*
*  24 Aout 1995 : WP : Rename from SafeString
*                      Adding << and >> operators .....
*                      we decided private inheritance from strstream
*                      protecting internal buffer.
*                      Move long inline methodes after class definition.
*                      Remove StrLen method.
*                      Rewrite SafeString::methods
*                      (new implementation based in  strstream).
*
*  25 Aout 1995 : WP : add == operator
*
*  1 November 1995 : Jan Madsen : reworked implementation.
*
************************************************************************/

#ifndef _CHAR_STRING_INC
#define _CHAR_STRING_INC

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

class CharString {
public:
  // ------------------------------------------------------------
  // constructor, copy str in memory allocated internally.

  CharString(const char *str=0);

  // ------------------------------------------------------------
  // constructor, Allocate memory and copy the strstream::Buffer() in it.

  CharString(const CharString &cs);
  
  // ------------------------------------------------------------
  // will destroy the memory allocated internally.

  virtual ~CharString() { if (buffer != NULL) free(buffer); }

  // ------------------------------------------------------------
  // Replace current contents with str.
  // Next extraction will be always possible (if str is not 0).
  virtual void Copy(const char *str);

  // ------------------------------------------------------------
  // Append a string

  virtual void Append (const char *str);
  
  // ------------------------------------------------------------
  // Append a character

  virtual void Append (char c);
  
  // ------------------------------------------------------------
  // Return a pointer to internal allocated memory.

  virtual const char *Buffer() const { if (buffer == NULL)
                                          return "\0";
                                       else
                                          return buffer; }

  // ------------------------------------------------------------
  // use a CharString where "const char *" can be used :

  operator const char *() { if (buffer == NULL)
                               return "\0";
                            else
                               return buffer; }

  // ------------------------------------------------------------
  // Get string length

  virtual int StrLen() const { if (buffer == NULL)
                                  return 0;
                               return strlen(buffer); }

  // ------------------------------------------------------------
  // = operator.
  CharString & operator=(const CharString &cs);

  // ------------------------------------------------------------
  // == operator (return value as strcmp fuction).
  int operator==(const CharString &cs) const;

  // Old version : free the memory allocated internally.
  // current version : Empty internal store. Next insertion will init at begin.
  virtual void Clean();

  // Converts the parameter to a string (if not already a string) and appends
  // it to the string inside the CharString object.
  CharString &operator << (char c);
  CharString &operator << (const char *str);
  CharString &operator << (const void *p);
  CharString &operator << (int i);
  CharString &operator << (long l);
  CharString &operator << (long long ll);
  CharString &operator << (short s);
  CharString &operator << (float f);
  CharString &operator << (double d);

  private :
    char *buffer;
    char *scanningPos; // Points in the string buffer to the first unread
                 // character. If buffer is NULL then scanningPos is also NULL.
};

inline
void CharString::Clean() {
   if (buffer != NULL)
      free(buffer);
   buffer = NULL;
}

inline
void CharString::Copy(const char *str) {
  int i, len;
  Clean();
  if (str != 0) {
     len = strlen(str);
     buffer = (char *) malloc(len + 1);
     if (buffer == NULL) {
        printf("CharString error: can't allocate memory\n");
        return;
     };
     for (i = 0; i < len; i++)
        buffer[i] = str[i];
     buffer[len] = '\0';
  };
}

inline
void CharString::Append(const char *str) {
  if (str != NULL) {
     if (buffer == NULL)
        Copy(str);
     else
     {
        char * newBuffer = (char *) malloc(strlen(buffer) + strlen(str) +1);
        int i;
        if (newBuffer == NULL) {
           printf("CharString error: can't allocate string memory.\n");
	   return ;
	};
        for (i = 0; i < strlen(buffer); i++)
           newBuffer[i] = buffer[i];
        for (i = 0; i < strlen(str); i++)
           newBuffer[i + strlen(buffer)] = str[i];
        newBuffer[strlen(buffer) + strlen(str)] = '\0';
        free(buffer);
        buffer = newBuffer;
        newBuffer = NULL;
     };
  };
}

inline
void CharString::Append(char c) {
  if (buffer == NULL) {
     buffer = (char *) malloc(2);
     buffer[0] = c;
     buffer[1] = '\0';
  }
  else
  {
     char * newBuffer = (char *) malloc(strlen(buffer) + 2);
     int i;
     if (newBuffer == NULL) {
        printf("CharString error: can't allocate string memory.\n");
	return ;
     };
     for (i = 0; i < strlen(buffer); i++)
        newBuffer[i] = buffer[i];
     newBuffer[strlen(buffer)] = c;
     newBuffer[strlen(buffer) + 1] = '\0';
     free(buffer);
     buffer = newBuffer;
     newBuffer = NULL;
  };
}

inline
CharString &CharString::operator=(const CharString &cs) {
  if ( this != &cs)
      Copy(cs.Buffer());
  return *this;
}

inline
int CharString::operator==(const CharString &cs) const {
  if (buffer == NULL && cs.StrLen() == 0)
     return 1;
  else if (buffer == NULL || cs.StrLen() == 0)
     return 0;
  else
     return !strcmp(buffer, cs.Buffer());
}

// Constructors (use inline methods) :
inline
CharString::CharString(const char *str=0) {
  buffer = NULL;
  Copy(str);
}

// ------------------------------------------------------------
// constructor, Allocate memory and copy the parameter's Buffer() in it.
inline
CharString::CharString(const CharString &cs) {
  buffer = NULL;
  Copy(cs.Buffer());
}

inline
CharString &CharString::operator << (char c)
{
   Append(c);
   return *this;
};

inline
CharString &CharString::operator << (const char *str)
{
   Append(str);
   return *this;
};

inline
CharString &CharString::operator << (const void *p)
{
   char buf[20];
   sprintf(buf, "%p\0", p);
   Append(buf);
   return *this;
};

inline
CharString &CharString::operator << (int i)
{
   char buf[10];
   sprintf(buf, "%i\0", i);
   Append(buf);
   return *this;
};

inline
CharString &CharString::operator << (long l)
{
   char buf[20];
   sprintf(buf, "%ld\0", l);
   Append(buf);
   return *this;
};

inline
CharString &CharString::operator << (long long ll)
{
   char buf[40];
   sprintf(buf, "%ld\0", ll);
   Append(buf);
   return *this;
};

inline
CharString &CharString::operator << (short s)
{
   char buf[10];
   sprintf(buf, "%i\0", s);
   Append(buf);
   return *this;
};

inline
CharString &CharString::operator << (float f)
{
   char buf[30];
   sprintf(buf, "%f\0", f);
   Append(buf);
   return *this;
};

inline
CharString &CharString::operator << (double d)
{
   char buf[60];
   sprintf(buf, "%f\0", d);
   Append(buf);
   return *this;
};

#endif

