//------------------------------------------------------------------------
//   $Id$
//   descr:      template class DynamicListTemplate
//               to manage dynamic list of user items
//               that can be accessed as an array efficiently implemented
//
//   author:     DIVIN Abel
//
//   $Log$
//------------------------------------------------------------------------

#ifndef _DYNAMIC_Array_TEMPLATE_H
#define _DYNAMIC_Array_TEMPLATE_H

#include <ListTemplate.h>

template <class UserItem> class DynamicArrayTemplate : public ListTemplate<UserItem> {
public:
  
  // ------------------------------------------------------------
  // create an array of size firstSize of pointers to UserItem
      
  DynamicArrayTemplate (
    // start size of internal array
    int firstSize = 50,
    // increment size of array of incrementSize each time the array is full
    int incrementSize = 50
  );

  // ------------------------------------------------------------
  // will destroy the Array
  // WARNING: by default, deleting a non empty Array will NOT delete leaved user data
      
  ~DynamicArrayTemplate();

  // ------------------------------------------------------------
  // do not refer in array user data contained in item number specified in parameter
  // WILL NOT FREE ANY DATA
  // return values:
  //   0 success
  //   1 wrong item

  int DelItem(int itemNr);

  // ------------------------------------------------------------
  // Insert an item at specified index,
  // and move the rest of items by one position

  void Insert(const UserItem &m, int index);

  // ------------------------------------------------------------
  // give access to item in position i
      
  UserItem operator [](int i) const;

  // ------------------------------------------------------------
  // Add an Item at the end of the Array

  virtual void Append(const UserItem &data);

  // ------------------------------------------------------------
  // Add an Item at the beginning of the Array

  virtual void Prepend(const UserItem &data);

  // ------------------------------------------------------------
  // return next item from Array without extracting it
  // and automatically move to next item in array
  // Exception is fired in case of :
  //   access to an empty array
  //   end of array already reached

  virtual UserItem ReadNext();


  // ------------------------------------------------------------
  // return 1 if end of array is reached
  //        0 if end of array is not reached

  virtual int EndOfList();


  // ------------------------------------------------------------
  // set current position in array to beginning of array

  virtual void Reset();


  // ------------------------------------------------------------
  // return 1 if Array is empty, 0 otherwise

  virtual int IsEmpty() const;


  // ------------------------------------------------------------
  // return number of items in array

  virtual int Size() const;


  // ------------------------------------------------------------
  // empty the array
  // WILL NOT FREE ANY DATA

  virtual void Clear();

private:

  const int DEFAULT_SIZE = 50;

  int incrementSize;
  int currentSize;
  int maxSize;
  int currentPos;
  UserItem *array;

  void CheckForRealloc();
};

template<class UserItem>
inline DynamicArrayTemplate<UserItem>::DynamicArrayTemplate (int firstSize, int incrementSize) {
  currentSize = 0; currentPos = -1;
  this->incrementSize = incrementSize;
  if (firstSize <= 0)
      maxSize = DEFAULT_SIZE;
  else
      maxSize = firstSize;

  array = new UserItem [firstSize];
}

template <class UserItem>
inline DynamicArrayTemplate<UserItem>::~DynamicArrayTemplate() {
  Clear();
  if (array != NULL)
    delete [] array;
}

template <class UserItem>
inline  int DynamicArrayTemplate<UserItem>::EndOfList() {
  return (currentPos >= currentSize)||(currentPos < 0);
}

template <class UserItem>
inline  void DynamicArrayTemplate<UserItem>::Reset() {
  currentPos = 0;
}

template <class UserItem>
inline int DynamicArrayTemplate<UserItem>::IsEmpty() const {
  return ( currentSize == 0 );
}

template <class UserItem>
inline  int DynamicArrayTemplate<UserItem>::Size() const {
  return currentSize;
}

template <class UserItem>
inline int DynamicArrayTemplate<UserItem>::DelItem(int itemNr) {
  if (itemNr >= 0 && itemNr < currentSize) {
    // move from that position the rest of items by one position
    for (int i = itemNr; i < currentSize-1; i++) {
      array[i] = array[i + 1];
    }
    
    currentSize--;
    currentPos = -1;
    
    return 0;
  }
  // wrong item number
  else { 
    return 1;
  }
}

template<class UserItem>
inline void DynamicArrayTemplate<UserItem>::Append(const UserItem &item) {
  array[currentSize] = item;
  currentSize++;
  currentPos = -1;
  CheckForRealloc();
}

template<class UserItem>
inline void DynamicArrayTemplate<UserItem>::Prepend(const UserItem &item) {
  for (int i = currentSize-1; i >= 0; i--) {
    array[i+1] = array[i];
  }
  
  array[0] = item;
  currentSize++;
  currentPos = -1;
  CheckForRealloc();
}

template<class UserItem>
inline void DynamicArrayTemplate<UserItem>::Insert(const UserItem &item, int position) {
  for (int i = currentSize-1; i >= position; i--) {
    array[i+1] = array[i];
  }
  
  array[position] = item;
  currentSize++;
  currentPos = -1;
  CheckForRealloc();
}

template<class UserItem>
inline  UserItem DynamicArrayTemplate<UserItem>::operator [](int i) const{
  return array[i];
}

template<class UserItem>
inline  void DynamicArrayTemplate<UserItem>::Clear() {
  currentSize = 0;
}

template<class UserItem>
inline void DynamicArrayTemplate<UserItem>::CheckForRealloc () {
  if (currentSize >= maxSize) {
    UserItem *tmpArray = new UserItem [maxSize + incrementSize];
    for(int i = 0; i < currentSize; i++) tmpArray[i] = array[i];
    delete [] array;
    array = tmpArray;
    maxSize += incrementSize;
  }
}

template<class UserItem>
inline  UserItem DynamicArrayTemplate<UserItem>::ReadNext() {
  currentPos++;
  return array[currentPos-1];
}

#endif
