//------------------------------------------------------------------------
//   $Id$
//   descr:      template class DynamicListTemplate
//               to manage dynamic list of copy of user items
//
//               A current item position in list is managed.
//
//   author:     DIVIN Abel
//
//   $Log$
//------------------------------------------------------------------------

#ifndef _DYNAMIC_LIST_TEMPLATE_H
#define _DYNAMIC_LIST_TEMPLATE_H

#include <ListTemplate.h>

template <class UserItem> class DynamicListTemplate : public ListTemplate<UserItem> {
public:

  // ------------------------------------------------------------
  // constructor, create an empty list

  DynamicListTemplate () { oldItem = NULL; }


  // ------------------------------------------------------------
  // will destroy the list

  ~DynamicListTemplate() { DeleteOldItem(); }


  // ------------------------------------------------------------
  // set current position in list to beginning of list
  // Return values:
  //   0 success
  //   1 failure because list is empty

  virtual int MoveToFirst();


  // ------------------------------------------------------------
  // set current position in list to end of list
  // Return values:
  //   0 success
  //   1 failure because list is empty

  virtual int MoveToLast();


  // ------------------------------------------------------------
  // set current position in list to next item in list
  // Return values:
  //   0 success
  //   1 failure because list is empty or end of list reached

  virtual int MoveToNext();


  // ------------------------------------------------------------
  // set current position in list to previous item in list
  // Return values:
  //   0 success
  //   1 failure because list is empty or end of list reached

  virtual int MoveToPrevious();


  // ------------------------------------------------------------
  // return current item from List without extracting it
  // Exception is fired in case of :
  //   access to an empty list
  //   end of list already reached

  virtual UserItem ReadCurrent();


  // ------------------------------------------------------------
  // Extract current Item from List, and move current to next in the list
  // Exception is fired in case of :
  //   access to an empty list
  //   end of list already reached

  virtual UserItem ExtractCurrent();


  // ------------------------------------------------------------
  // Insert userItem in list just before item pointed by Current.
  // Current position is unchanged
  // Exception is fired in case of :
  //   access to an empty list
  //   end of list already reached

  virtual void InsertBeforeCurrent(const UserItem &userItem);


  // ------------------------------------------------------------
  // Insert userItem in list just after item pointed by Current.
  // Current position is unchanged.
  // Exception is fired in case of :
  //   access to an empty list
  //   end of list already reached

  virtual void InsertAfterCurrent(const UserItem &userItem);


private:
  // if there exists an old extracted item to be deleted
  // WHY This item cannot be deleted when extracted ?
  // Because copy to user is done only during return from ExtractCurrent()
  Item *oldItem;

  void DeleteOldItem();
};

template <class UserItem>
int DynamicListTemplate<UserItem>::MoveToFirst() {
  current = first;
  return (current == NULL);
}

template <class UserItem>
int DynamicListTemplate<UserItem>::MoveToLast() {
  current = last;
  return (current == NULL);
}

template <class UserItem>
int DynamicListTemplate<UserItem>::MoveToNext() {
  if (current != NULL) current = current->next;
  return (current == NULL);
}

template <class UserItem>
int DynamicListTemplate<UserItem>::MoveToPrevious() {
  if (current != NULL) current = current->previous;
  return (current == NULL);
}

template <class UserItem>
UserItem DynamicListTemplate<UserItem>::ReadCurrent() {
  // empty List
  if (first == NULL) {
    ThrowException("DynamicListTemplate.ReadCurrent", "CANNOT perform action on empty list");
  }
  // current invalid position
  else if (current == NULL) {
    ThrowException("DynamicListTemplate.ReadCurrent",
		  "CANNOT perform action because end of list reached");
  }
  // valid current position
  else {
    return ( current->userData );
  }
}

template <class UserItem>
UserItem DynamicListTemplate<UserItem>::ExtractCurrent() {
  Item *item;

  // check if a previous extracted old item exists, and delete it
  DeleteOldItem();

  // empty List
  if (first == NULL) {
    ThrowException("DynamicListTemplate.ExtractCurrent",
		  "CANNOT perform action on empty list\n");
  }
  // current invalid position
  else if (current == NULL) {
    ThrowException("DynamicListTemplate.ExtractCurrent",
		  "CANNOT perform action because end of list reached\n");
  }
  // valid current position
  else {
    NrOfItems--;
    item = current;

    // current is on first item
    if (current == first) {
      first = current = first->next;
      if (first == NULL) last = NULL;
    }
    // current is on last item and not on first => number of items >= 2
    else if (current == last) {
      last = last->previous;
      last->next = NULL;
      current = NULL;
    }
    // current is not on first and not on last item => number of items > 2
    else {
      current->previous->next = current->next;
      current->next->previous = current->previous;

      current = current->next;
    }

    // will be delete during next call to ExtractCurrent
    // because copy from item to user is performed only during return
    oldItem = item;

    return ( item->userData );
  }
}

template <class UserItem>
void DynamicListTemplate<UserItem>::InsertBeforeCurrent
(
 const UserItem &userItem
 ) {
  Item *item;
  
  // empty List
  if (first == NULL) {
    ThrowException("DynamicListTemplate.InsertBeforeCurrent",
		   "CANNOT perform action on empty list\n");
  }
  // current invalid position
  else if (current == NULL) {
    ThrowException("DynamicListTemplate.InsertBeforeCurrent",
		   "CANNOT perform action because end of list reached\n");
  }
  // valid current position
  else {
    NrOfItems++;
    item = new Item(userItem);
    
    if (current == first) {
      // chaining new item
      item->next = first;
      first->previous = item;
      
      // update first
      first = item;
    }
    else {
      // chaining item
      item->next = current;
      item->previous = current->previous;

      // update previous item
      current->previous->next = item;

      // update current item
      current->previous = item;

    }
    
  }
}

template <class UserItem>
void DynamicListTemplate<UserItem>::InsertAfterCurrent
(
 const UserItem &userItem
 ) {
  Item *item;

  // empty List
  if (first == NULL) {
    ThrowException("DynamicListTemplate.InsertAfterCurrent",
		  "CANNOT perform action on empty list\n");
  }
  // current invalid position
  else if (current == NULL) {
    ThrowException("DynamicListTemplate.InsertAfterCurrent",
		  "CANNOT perform action because end of list reached\n");
  }
  // valid current position
  else {
    NrOfItems++;
    item = new Item(userItem);

    if (current == last) {
      // chaining new item
      last->next = item;
      item->previous = last;

      // update last
      last = item;
    }
    else {
      // chaining item
      item->next = current->next;
      item->previous = current;

      // update next item
      current->next->previous = item;

      // update current item
      current->next = item;
    }

  }
}

template <class UserItem>
void DynamicListTemplate<UserItem>::DeleteOldItem() {
  if (oldItem != NULL) {
    delete oldItem;
    oldItem = NULL;
  }
}

#endif
