/*
 * Synchronous C++ include file for any Synchronous C++ appplication
 * Copyright: EPFL - LABORATOIRE DE TELEINFORMATIQUE
 * Author: Divin Abel / Madsen Jan
 */

#ifndef _SCXX_USER_INC
#define _SCXX_USER_INC

#ifdef __cplusplus
extern "C++" {
#endif

// -----------------    Type declarations   --------------------

struct SystemType {
  typedef int FileDescr;
  enum Mode { RMODE, WMODE, XMODE};
};

extern int __SCXX_DEFAULT_STACK_SIZE;     // Synchronous C++ default coroutine stack size


// -----------------     User->Kernel interface ----------------

// Allow the debugging with the XDebugger.
extern int SetXDebug(char *hostName, int port);

//-----------------------------------------------------------------------
// sets debugging output for a single active object. If set to 1, debugging
// output is sent to console. Can be set and reset during execution.
// Can be called only inside an active class.

#define SetDebug(on) { __SCXX_SET_DEBUG(this->__SCXX_BP,on); }

//-----------------------------------------------------------------------
// sets debugging output for main program. If set to 1, debugging
// output is sent to console. Can be set and reset during execution.
//

extern void SetMainDebug(int on);


//-----------------------------------------------------------------------
// sets global debugging. If >= 1 debugging output is sent to console.
// 0 <= debug level <= 9
//   0     means no debug
//   1     set output messages for constructors and start of bodies
//   2     set ouput messages for context switches
//   3     set output messages for all user active objects synchronizations
//   4     set output messages for external events and unix sockets/unix gates
//   5,6,7 set debug on for all Synchronous C++ librairies
//   8,9   set kernel debug
// Can be set and reset during execution.
//

extern void SetKernelDebug(int level);


//-----------------------------------------------------------------------
// returns the debug level set for all Synchronous C++ librairies by the
// user using SetKernelDebug() to a value from 5 to 7.
//   0 means no debug set.
//   0 <= debug level <= 3

extern int GetLibDebugLevel();


//-----------------------------------------------------------------------
// Returns the time, which is the number of "clock ticks" since the program
// started execution. By default, a clock tick occurs 100 times a second but
// can be modified by SetResolution.
//

extern unsigned long (*now) (void);


//-----------------------------------------------------------------------
// used to initialize the kernel in virtual time mode
//

extern int SetVirtualTimeMode(void);


//-----------------------------------------------------------------------
// automatically called by kernel library,
// used to initialize the kernel in real time mode
//

extern int SetRealTimeMode(void);


//-----------------------------------------------------------------------
// exit program with exit_value, to be used instead of exit on IBM PC
//

extern void StopAll(int exit_value);


//-----------------------------------------------------------------------
// terminates the main task but not the application.
// To be called only by main itself
//

extern void StopMain();


/*
 * -=-=-=-=-=-=-=-=-=-==-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 * Compiler-Preprocessor -> Kernel interface
 * Procedures that only code inserted by the preprocessor uses.
 * -=-=-=-=-=-=-=-=-=-==-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
 */

//-----------------------------------------------------------------------
// Initialize the __SCXX_BP variable of active class by
// allocating first a blocprocess allocated before initializing it
// Always return 0;
// Added at the beginning of any active class constructor
//

extern short __SCXX_INIT_BP (
  char * className,     // string containing the name of the active class.
  void ** ptrToBpPtr,   // pointer to the active class blocprocess variable __SCXX_BP.
  void * thisPtr        // this pointer value of the instance of the active class, used
                        // in debugging for distinguishing between different active
                        // objects, they have no effect on the running of the process.
);

//-----------------------------------------------------------------------
// set debug on/off for a particular bloc process
//

extern void __SCXX_SET_DEBUG (
  void * bpPtr,          // pointer to a bloc process
  int debug              // 1: debug on, 0: debug off
);

//---------------------------------------------------------------------------------
// Creates a synchronization point (also known as: "Mail Box" or non virtual "Gate")
// Always return 0;
// Added at the beginning of any constructor of an active class

extern short __SCXX_CREATE_GATE (
  void * bpPtr,          // pointer to active object bloc process containing this call
  char * methodName,     // method name
  int methodNr,          // method number, 0 means destructor
  void * methodAddr=0    // address of the method, 0 in this version
);

//--------------------------------------------------------------------------------------
// create a multiple synchronization point (also known as: "Mail Box" or "virtual Gate")
// Always return 0;
// Added at the beginning of any constructor of an active class
//

extern short __SCXX_CREATE_VIRTUAL_GATE (
  void * bpPtr,       // pointer to active object bloc process containing this call
  char * methodName,  // method name
  int methodNr,       // method number != 0
  void * methodAddr=0 // address of the method, 0 in this version
);

//------------------------------------------------------------------------------
// Add one more active class to participation in a virtual gate synchronization
// by modifying base class gates to the newly created one
// Always return 0;
// Added at the beginning of any constructor of an active class
//

extern short __SCXX_INCREMENT_VIRTUAL_GATE (
  void * bpPtr,          // pointer to active object bloc process containing this call
  int methodNr,          // method number != 0
  void *baseClassBpPtr,  // pointer to a base active class bloc process that knows
                         // the method to be virtual
  int baseClassMethodNr  // method number related to the base class != 0
);

//-----------------------------------------------------------------------
// Starts a parameterless procedure body (known as the body of an active class),
// running as a process.
// Added at the end of any active class constructor
//

extern void __SCXX_START_BODY (
  void *bpPtr,           // pointer to active object bloc process containing this call
  void (*body)(),        // pointer on active class body method
  int stackSize          // stack size, default __SCXX_DEFAULT_STACK_SIZE
);

//-----------------------------------------------------------------------
//  return this pointer for current active object in execution

extern void *__SCXX_GET_THIS_PTR ();

//-----------------------------------------------------------------------
// Added at the beginning of the destructor of an active class
//

extern short __SCXX_CALL_DESTRUCTOR (
  void *bpPtr           // pointer to active object bloc process containing this call
);

//-----------------------------------------------------------------------
// Added before 'delete activeObject' inside a select only, to disable
// asking kernel to synchronize on ~Gate by code placed at the beginning
// the destructor
//

extern void __SCXX_DISABLE_SYNC_DESTRUCTOR (
  void *calledBpPtr    // pointer to bloc process of called active object
);

//-----------------------------------------------------------------------
// stop body execution, kill coroutine and free blocprocess and gates
// Added at the end of any active class destructor

extern void __SCXX_STOP_AND_KILL_BODY (
  void **ptrToBPPtr      // pointer to the active class blocprocess variable __SCXX_BP
);

//-----------------------------------------------------------------------
// enters an infinite loop over a select statement on all methods
// of the current active object
// Added at the end of the body and correspond to the default body execution

extern void __SCXX_DEFAULT_BODY();

/*
 * -----------------     Synchronization calls   ----------------
 *
 * to each public or protected method of an active object is associated
 * a unique number given by the compiler. This number is always the same
 * for each instance of a same active class. Numbers follow by one,
 * from 0 to NbGates-1, NbGates represents the number of declared gates.
 * All destructor gates have number 0 and are non virtual gates.
 *
 */

//-----------------------------------------------------------------------
// Replace statement 'accept method;' outside a select
//

extern void __SCXX_ACCEPT (
  void *bpPtr,         // pointer to active object bloc process containing this call
  int methodNr         // method number called
);

//------------------------------------------------------------------------
// Replace statement 'accept method from object ;' outside a select
//

extern void __SCXX_ACCEPT_FROM (
 void *bpPtr,		// pointer to active object bloc process containing this call
 int methodNr,		// method number called
 void *fromPtr		 //pointer to object to accept this call from
);

//-----------------------------------------------------------------------
// Added before statement 'activeObject.method(...);' ouside a select
// always return 1

extern short __SCXX_CALL (
  void *calledBpPtr,   // pointer to bloc process of called active object
  int methodNr         // method number called
);

//-----------------------------------------------------------------------
// Will put on the ready list the active object that synchronizes with accept
// Added after statement 'activeObject.method(...);'
// even if the call is in a select, always return 0
//

extern unsigned int __SCXX_RELEASE ();

//-----------------------------------------------------------------------
// Replace statement 'waituntil(time);' outside a select
//

extern void __SCXX_WAITUNTIL(
  unsigned long time   // absolute time to wait for
);

//-----------------------------------------------------------------------
//
// Added before statement 'activeObject.method(...);' inside a select
//

extern void __SCXX_SELECT_CALL (
  void *calledBpPtr,   // Pointer to bloc process of called active object
  int methodNr,        // Method number called
  int alternativeNr    // Number to be returned by __SCXX_SYNC_REQUEST given by
                       // the compiler designating alternative number in the select
);

//-----------------------------------------------------------------------
//
// Added before statement 'activeObject(i).method(...);' inside a select
//

extern void __SCXX_SELECT_FORALL_CALL (
  void *calledBpPtr,   // Pointer to bloc process of called active object
  int methodNr,        // Method number called
  int alternativeNr,   // Number to be returned by __SCXX_SYNC_REQUEST given by
                       // the compiler designating alternative number in the select
  int index            // index used in forall statement
);

//-----------------------------------------------------------------------
// Replace statement 'accept method' inside a select
//

extern void __SCXX_SELECT_ACCEPT (
  void *bpPtr,         // pointer to active object bloc process containing this call
  int methodNr,        // Method number called
  int alternativeNr    // Number to be returned by __SCXX_SYNC_REQUEST given by
                       // the compiler designating alternative number in the select
);

//-----------------------------------------------------------------------
// Replace statement 'accept method from object' inside a select
//

extern void __SCXX_SELECT_ACCEPT_FROM (
  void *bpPtr,         // pointer to active object bloc process containing this call
  int methodNr,        // Method number called
  int alternativeNr,    // Number to be returned by __SCXX_SYNC_REQUEST given by
                       // the compiler designating alternative number in the select
  void *fromPtr        // pointer to active object bloc process to accept this call from
);

//-----------------------------------------------------------------------

// Replace statement 'waituntil(time);' inside a select
//

extern void __SCXX_SELECT_WAITUNTIL (
  unsigned long time,  // absolute time to wait for
  int alternativeNr    // Number to be returned by __SCXX_SYNC_REQUEST given by
                       // the compiler designating alternative number in the select
);

//-----------------------------------------------------------------------
// Replace 'default' clause inside a select
//

extern void __SCXX_SELECT_DEFAULT (
  int alternativeNr    // Number to be returned by __SCXX_SYNC_REQUEST given by
                       // the compiler designating alternative number in the select
);

//-----------------------------------------------------------------------
// Returns the index of the alternative selected by the kernel
// Added in the coding of a select to wait for synchronization on an
// event list
//

extern int __SCXX_SYNC_REQUEST ();

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
// Returns the index of the alternative selected by the kernel
// Added in the coding of a select to wait for synchronization on an
// event list
//
//

extern int __SCXX_SYNC_REQUEST (void *fromPtr);

//-----------------------------------------------------------------------

// Returns the index of the alternative selected by the kernel for a
// specific forall statement.
// Added in the coding of a select to return index value of the choosen
// synchronization
//

extern int __SCXX_FORALL_INDEX ();

// This change is due to the fact that the names 'accept' and 'select' which
// is reserved words in sC++ also exists as command names under UNIX.
// We hereby change the names to __SCXX_UNIX_ACCCEPT and __SCXX_UNIX_SELECT. 
#define accept __SCXX_UNIX_ACCEPT
#define select __SCXX_UNIX_SELECT
#include <unistd.h>
#include <sys/socket.h>
#include <sys/types.h>
#undef accept
#undef select

#ifdef __cplusplus
};
#endif

#endif /* _SCXX_USER_INC */

